import copy
import json
from typing import Dict, List, Type, Union, cast

from gentrace.api_client import ApiClient
from gentrace.apis.tags.ingestion_api import IngestionApi
from gentrace.configuration import Configuration
from gentrace.providers.llms.openai import OpenAIPipelineHandler
from gentrace.providers.step_run import StepRun


class PipelineRun:
    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.step_runs: List[StepRun] = []

    def get_pipeline(self):
        return self.pipeline

    def get_openai(self):
        if "openai" in self.pipeline.pipeline_handlers:
            handler = self.pipeline.pipeline_handlers.get("openai")
            cloned_handler = copy.deepcopy(handler)
            import openai

            from .llms.openai import (
                intercept_chat_completion,
                intercept_completion,
                intercept_embedding,
            )

            for name, cls in vars(openai.api_resources).items():
                if isinstance(cls, type):
                    # Create new class that inherits from the original class, don't directly monkey patch
                    # the original class
                    new_class = type(name, (cls,), {})
                    if name == "Completion":
                        new_class.create = intercept_completion(new_class.create)
                    elif name == "ChatCompletion":
                        new_class.create = intercept_chat_completion(new_class.create)
                    elif name == "Embedding":
                        new_class.create = intercept_embedding(new_class.create)

                    new_class.pipeline_run = self

                    setattr(cloned_handler, name, new_class)
                    # TODO: must work on a acreate() method, check that streaming works

            cloned_handler.set_pipeline_run(self)

            # TODO: Could not find an easy way to create a union type with openai and
            # OpenAIPipelineHandler, so we just use openai.
            return cast(openai, cloned_handler)
        else:
            raise ValueError(
                "Did not find OpenAI handler. Did you call setup() on the pipeline?"
            )

    def get_pinecone(self):
        if "pinecone" in self.pipeline.pipeline_handlers:
            handler = self.pipeline.pipeline_handlers.get("pinecone")
            cloned_handler = copy.deepcopy(handler)
            cloned_handler.set_pipeline_run(self)
            import pinecone

            return cast(pinecone, cloned_handler)
        else:
            raise ValueError(
                "Did not find Pinecone handler. Did you call setup() on the pipeline?"
            )

    def add_step_run(self, step_run: StepRun):
        self.step_runs.append(step_run)

    def submit(self) -> Dict:
        configuration = Configuration(host=self.pipeline.config.get("host"))
        configuration.access_token = self.pipeline.config.get("api_key")
        api_client = ApiClient(configuration=configuration)
        ingestion_api = IngestionApi(api_client=api_client)

        step_runs_data = [
            {
                "provider": {
                    "name": step_run.provider,
                    "invocation": step_run.invocation,
                    "modelParams": step_run.model_params,
                    "inputs": step_run.inputs,
                    "outputs": step_run.outputs,
                },
                "elapsedTime": step_run.elapsed_time,
                "startTime": step_run.start_time,
                "endTime": step_run.end_time,
            }
            for step_run in self.step_runs
        ]

        print("Step run: ", json.dumps(step_runs_data, indent=2))

        pipeline_post_response = ingestion_api.pipeline_run_post(
            {"name": self.pipeline.id, "stepRuns": step_runs_data}
        )

        return {
            "pipelineRunId": pipeline_post_response.body.get_item_oapg("pipelineRunId")
        }



import time
from datetime import datetime
from typing import Optional

import openai
import pystache

from gentrace.providers.step_run import StepRun
from gentrace.providers.utils import to_date_string


class OpenAIPipelineHandler:
    def __init__(self, pipeline=None):
        self.pipeline = pipeline

    @classmethod
    def setup(cls, config):
        if config:
            for key, value in config.items():
                setattr(openai, key, value)

    def set_pipeline_run(self, pipeline_run):
        self.pipeline_run = pipeline_run


def intercept_completion(original_fn):
    @classmethod
    def wrapper(cls, *args, **kwargs):
        prompt_template = kwargs.get("promptTemplate")
        prompt_inputs = kwargs.get("promptInputs")
        base_completion_options = {
            k: v
            for k, v in kwargs.items()
            if k not in ["promptTemplate", "promptInputs"]
        }

        if "prompt" in base_completion_options:
            raise ValueError(
                "The prompt attribute cannot be provided when using the Gentrace SDK. Use promptTemplate and promptInputs instead."
            )

        if not prompt_template:
            raise ValueError(
                "The promptTemplate attribute must be provided when using the Gentrace SDK."
            )

        rendered_prompt = pystache.render(prompt_template, prompt_inputs)

        new_completion_options = {**base_completion_options, "prompt": rendered_prompt}

        start_time = time.time()
        completion = original_fn(**new_completion_options)
        end_time = time.time()

        elapsed_time = int(end_time - start_time)

        user = base_completion_options.get("user")
        suffix = base_completion_options.get("suffix")
        partial_model_params = {
            k: v
            for k, v in base_completion_options.items()
            if k not in ["user", "suffix"]
        }

        inputs_dict = {"prompt": prompt_inputs}
        if user is not None:
            inputs_dict["user"] = user
        if suffix is not None:
            inputs_dict["suffix"] = suffix

        cls.pipeline_run.add_step_run(
            OpenAICreateCompletionStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                inputs_dict,
                {**partial_model_params, "promptTemplate": prompt_template},
                completion,
            )
        )
        return completion

    return wrapper


def intercept_chat_completion(original_fn):
    @classmethod
    def wrapper(cls, *args, **kwargs):
        messages = kwargs.get("messages")
        user = kwargs.get("user")
        model_params = {
            k: v for k, v in kwargs.items() if k not in ["messages", "user"]
        }

        start_time = time.time()
        completion = original_fn(**kwargs)
        end_time = time.time()

        elapsed_time = int(end_time - start_time)

        cls.pipeline_run.add_step_run(
            OpenAICreateChatCompletionStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                {"messages": messages, "user": user},
                model_params,
                completion,
            )
        )
        return completion

    return wrapper


def intercept_embedding(original_fn):
    @classmethod
    def wrapper(cls, *args, **kwargs):
        model = kwargs.get("model")
        input_params = {k: v for k, v in kwargs.items() if k not in ["model"]}

        start_time = time.time()
        completion = original_fn(**kwargs)
        end_time = time.time()

        elapsed_time = int(end_time - start_time)

        cls.pipeline_run.add_step_run(
            OpenAICreateEmbeddingStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                input_params,
                {"model": model},
                completion,
            )
        )
        return completion

    return wrapper


class OpenAICreateCompletionStepRun(StepRun):
    def __init__(
        self, elapsed_time, start_time, end_time, inputs, model_params, response
    ):
        super().__init__(
            "openai",
            "openai_createCompletion",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
        )
        self.inputs = inputs
        self.model_params = model_params
        self.response = response


class OpenAICreateChatCompletionStepRun(StepRun):
    def __init__(
        self, elapsed_time, start_time, end_time, inputs, model_params, response
    ):
        super().__init__(
            "openai",
            "openai_createChatCompletion",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
        )
        self.inputs = inputs
        self.model_params = model_params
        self.response = response


class OpenAICreateEmbeddingStepRun(StepRun):
    def __init__(
        self, elapsed_time, start_time, end_time, inputs, model_params, response
    ):
        super().__init__(
            "openai",
            "openai_createEmbedding",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
        )
        self.inputs = inputs
        self.model_params = model_params
        self.response = response
